(function() {
    "use strict";

    angular.module("weChatApp")
        .factory("PermissionService", ["$q", "Endpoints", PermissionService])
    ;

    function PermissionService($q, endpoints) {
        var service = this;

        /**
         * Check if user have access to the given object list.
         * If objectList is omitted, check against the objects defined within Queries
         * @param queryObjects the map of query object
         * @returns {*}
         */
        service.checkObjectFLS = function(queryObjects) {
            var deferred = $q.defer();

            endpoints.fetchAvailableObjects().then(function(resp) {
                var availableObjects = resp.data;
                var unreachableObjects = {};

                for(var key in queryObjects) {
                    if(Object.prototype.hasOwnProperty.call(queryObjects, key)) {
                        var objName = queryObjects[key].object;
                        if(!availableObjects[objName]) {
                            unreachableObjects[objName] = [];
                        }
                    }
                }
                var unreachableObjectsList = _.keys(unreachableObjects);
                if(unreachableObjectsList.length > 0) {
                    console.error("Cannot read object: " + unreachableObjectsList.join(", "));
                }
                deferred.resolve(unreachableObjects);
            }, function(error) {
                deferred.reject(error);
            });

            return deferred.promise;
        };

        /**
         * Check if user have access to the given fields under the given object.
         * If both parameters are omitted, check against the objects and fields defined within Queries
         * @param queryObjects the map of query object
         * @returns {*}
         */
        service.checkFieldFLS = function(queryObjects) {
            var deferred = $q.defer();
            var checkFieldPromises = [];

            for(var key in queryObjects) {
                if(Object.prototype.hasOwnProperty.call(queryObjects, key)) {
                    var subDeferred = $q.defer();
                    checkFieldFlsHelper(queryObjects[key].object, queryObjects[key].fields, subDeferred);
                    checkFieldPromises.push(subDeferred.promise);
                }
            }

            $q.all(checkFieldPromises).then(function(comboResp) {
                var unreachableFieldsByObject = {};
                comboResp.forEach(function(resp) {
                    if(resp.fields && resp.object && !unreachableFieldsByObject[resp.object]) {
                        unreachableFieldsByObject[resp.object] = resp.fields;
                    }
                });
                deferred.resolve(unreachableFieldsByObject);
            }, function(error) {
                deferred.reject(error);
            });

            return deferred.promise;
        };

        /**
         * Check FLS for object and fields defined within Queries
         */
        service.checkObjectAndFieldFLS = function(queryObjects) {
            var deferred = $q.defer();

            service.checkObjectFLS(queryObjects).then(function(unreachableObjects) {
                var filteredQueryObjects = _.filter(queryObjects, function(qObj) {
                    return !angular.isDefined(unreachableObjects[qObj.object]);
                });
                service.checkFieldFLS(filteredQueryObjects).then(function(unreachableFieldsByObject) {
                    deferred.resolve(_.assign(unreachableObjects, unreachableFieldsByObject));
                }, function(error) {
                    deferred.reject(error);
                });
            }, function(error) {
                deferred.reject(error);
            });

            return deferred.promise;
        };

        function checkFieldFlsHelper(objectName, fieldList, deferred) {
            endpoints.fetchObjectMetadata(objectName).then(function(resp) {
                var availableFields = _.map(resp.data.fields, "name");
                var unknownFields = _.difference(fieldList, availableFields);
                var result = {};
                if(unknownFields.length > 0) {
                    console.error("Cannot read field: " + unknownFields.join(", ") + " from " + objectName);
                    result.object = objectName;
                    result.fields = unknownFields;
                }
                deferred.resolve(result);
            }, function(error) {
                deferred.reject(error);
            });
        }

        return service;
    }
}());
